from fqfa.constants import RNA_BASES
from mavehgvs.patterns.util import combine_patterns, remove_named_groups
from mavehgvs.patterns.position import pos_intron

rna_nt: str = rf"[{''.join(RNA_BASES).lower()}]"
"""str: Pattern matching any lowercase RNA base.

This does not include IUPAC ambiguity characters.
"""

rna_sub: str = rf"(?P<rna_sub>(?:(?P<position>{pos_intron})(?P<ref>{rna_nt})>(?P<new>{rna_nt}))|(?P<equal>=))"
"""str: Pattern matching a RNA substitution with numeric or relative-to-transcript positions.
"""

rna_del: str = rf"(?P<rna_del>(?:(?:(?P<start>{pos_intron})_(?P<end>{pos_intron}))|(?P<pos>{pos_intron}))del)"
"""str: Pattern matching a RNA deletion with numeric or relative-to-transcript positions.
"""

rna_dup: str = rf"(?P<rna_dup>(?:(?:(?P<start>{pos_intron})_(?P<end>{pos_intron})dup)|(?P<pos>{pos_intron}))dup)"
"""str: Pattern matching a RNA duplication with numeric or relative-to-transcript positions.
"""

rna_ins: str = rf"(?P<rna_ins>(?P<start>{pos_intron})_(?P<end>{pos_intron})ins(?P<seq>{rna_nt}+))"
"""str: Pattern matching a RNA insertion with numeric or relative-to-transcript positions.
"""

rna_delins: str = rf"(?P<rna_delins>(?:(?:(?P<start>{pos_intron})_(?P<end>{pos_intron}))|(?P<pos>{pos_intron}))delins(?P<seq>{rna_nt}+))"
"""str: Pattern matching a RNA deletion-insertion with numeric or relative-to-transcript positions.
"""

rna_variant: str = combine_patterns(
    [rna_sub, rna_del, rna_dup, rna_ins, rna_delins], None
)
"""str: Pattern matching any single RNA variant event.
"""

rna_single_variant: str = rf"(?P<rna>r\.{rna_variant})"
"""str: Pattern matching any complete RNA variant, including the prefix character.
"""

rna_multi_variant: str = rf"(?P<rna_multi>r\.\[{remove_named_groups(rna_variant)}(?:;{remove_named_groups(rna_variant)}){{1,}}\])"
"""str: Pattern matching any complete RNA multi-variant, including the prefix character.

Named capture groups have been removed from the variant patterns because of non-uniqueness.
Another applications of single-variant regular expressions is needed to recover the named groups from each individual
variant in the multi-variant.
"""



import re
from functools import total_ordering

from mavehgvs.exceptions import MaveHgvsParseError
from mavehgvs.patterns.position import pos
from mavehgvs.patterns.protein import amino_acid

__all__ = ["VariantPosition"]

pos_with_groups: str = rf"(?P<position_aa>{amino_acid})?(?P<position>[*-]?{pos})(?P<position_intron>[+-]{pos})?"
"""str: Pattern matching a position with match groups for parsing into a :py:class:`VariantPosition`.
"""


@total_ordering
class VariantPosition:
    """Class for storing a variant position.

    The class includes special fields for variants using the extended position syntax.
    Attributes
    ----------
    position : Optional[int]
        The position as an integer.
        Negative positions are only expected for 5' UTR positions.
    amino_acid : Optional[str]
         The amino acid at this position for protein variants.
    intronic_position : Optional[int]
        The number of bases into the intron for intronic positions.
        None for non-intronic positions.

        Nucleotides in the 5' half of the intron have positive ``intronic_position`` and their position is that of
        the last base of the 5' exon.
        Nucleotides in the 3' half of the intron have negative ``intronic_position`` and their position is that of
        the first base of the 3' exon.
    utr : Optional[bool]
        True if the position is in the UTR. None for all other positions.

    """

    fullmatch = re.compile(pos_with_groups, flags=re.ASCII).fullmatch
    """Callable[[str, int, int], Optional[Match[str]]]: fullmatch callable for parsing positions
    
    Returns an :py:obj:`re.Match` object if the full string matches one of the position groups in :py:data:`pos_extended`.
    """

    def __init__(self, pos_str: str) -> None:
        """Parse a position string into a VariantPosition object.

        Parameters
        ----------
        pos_str : str
            The string to convert to a VariantPosition object.

        """
        try:
            gdict = VariantPosition.fullmatch(pos_str).groupdict()
        except AttributeError:
            raise MaveHgvsParseError(f"invalid variant position string '{pos_str}'")

        self.position = None
        self.amino_acid = None
        self.intronic_position = None
        self.utr = None

        if gdict["position"].startswith("*"):  # 3' UTR position
            self.utr = True
            self.position = int(gdict["position"][1:])
        else:
            if gdict["position"].startswith("-"):  # 5' UTR position
                self.utr = True
            self.position = int(gdict["position"])

        if gdict["position_aa"] is not None:
            self.amino_acid = gdict["position_aa"]

        if gdict["position_intron"] is not None:
            self.intronic_position = int(gdict["position_intron"])

        if self.amino_acid is not None and (
            self.intronic_position is not None or self.utr is not None
        ):
            raise MaveHgvsParseError("invalid variant")

    def __repr__(self) -> str:
        """The object representation is equivalent to the input string.

        Returns
        -------
        str
            The object representation.

        """
        if self.utr and self.position > 0:
            p = f"*{self.position}"
        else:
            p = f"{self.position}"

        if self.intronic_position is not None:
            if self.intronic_position > 0:
                return f"{p}+{self.intronic_position}"
            else:
                return f"{p}{self.intronic_position}"
        elif self.amino_acid is not None:
            return f"{self.amino_acid}{p}"
        else:
            return p

    def __lt__(self, other: "VariantPosition") -> bool:
        """Less than comparison operator.

        Other comparison operators will be filled in using :py:func:`functools.total_ordering`.

        Parameters
        ----------
        other : VariantPosition
            The other VariantPosition to compare to.

        Returns
        -------
        bool
            True if this position evaluates as strictly less than the other position; else False.

        """
        if self.utr == other.utr:
            if self.position == other.position:
                if self.intronic_position == other.intronic_position:
                    return False
                elif self.intronic_position is None:
                    return other.intronic_position > 0
                elif other.intronic_position is None:
                    return self.intronic_position < 0
                else:
                    return self.intronic_position < other.intronic_position
            else:
                return self.position < other.position
        else:  # 5' < non-UTR < 3'
            if self.utr:
                if self.position < 0:  # self is in 5' UTR
                    return True
                else:  # self is in 3' UTR
                    return False
            else:
                if other.position < 0:  # other is in 5' UTR
                    return False
                else:  # other is in 3' UTR
                    return True

    def __eq__(self, other: "VariantPosition") -> bool:
        """Equality comparison operator.

        Note that the amino acid portion of a protein position is not used in this comparison.

        Other comparison operators will be filled in using :py:func:`functools.total_ordering`.

        Parameters
        ----------
        other : VariantPosition
            The other VariantPosition to compare to.

        Returns
        -------
        bool
            True if this position is the same as the other position; else False.

        """
        return (self.position, self.intronic_position, self.utr) == (
            other.position,
            other.intronic_position,
            other.utr,
        )

    def __ne__(self, other: "VariantPosition") -> bool:
        """Not equal comparison operator.

        Note that the amino acid portion of a protein position is not used in this comparison.

        Other comparison operators will be filled in using :py:func:`functools.total_ordering`.

        Parameters
        ----------
        other : VariantPosition
            The other VariantPosition to compare to.

        Returns
        -------
        bool
            True if this position is not the same as the other position; else False.

        """
        return (self.position, self.intronic_position, self.utr) != (
            other.position,
            other.intronic_position,
            other.utr,
        )

    def is_utr(self) -> bool:
        """Return whether this is a UTR position.

        Returns
        -------
        bool
            True if the object describes a position in the UTR; else False.

        """
        return self.utr is not None

    def is_intronic(self) -> bool:
        """Return whether this is an intronic position.

        Returns
        -------
        bool
            True if the object describes a position in an intron; else False.

        """
        return self.intronic_position is not None

    def is_protein(self) -> bool:
        """Return whether this is a protein position

        Returns
        -------
        bool
            True if the object describes a position with an amino acid component; else False.
        """
        return self.amino_acid is not None

    def is_extended(self) -> bool:
        """Return whether this position was described using the extended syntax.

        Returns
        -------
        bool
            True if the position was described using the extended syntax; else False.

        """
        return self.utr is not None or self.intronic_position is not None

    # the string annotation used in the type hint below is required for Python 3.6 compatibility
    def is_adjacent(self, other: "VariantPosition") -> bool:
        """Return whether this variant and another are immediately adjacent in sequence space.

        The following special cases are not handled correctly:

        * The special case involving the last variant in a transcript sequence and the first base in the 3'
          UTR will be evaluated as not adjacent, as the object does not have sequence length information.
        * The special case involving the two middle bases in an intron where the numbering switches from
          positive with respect to the 5' end of the intron to negative with respect to the 3' end of the intron
          will be evaluated as not adjacent, as the object does not have intron length information.
        * This ignores the special case where there is an intron between the last base of the 5' UTR and the first
          base of the coding sequence because it is not biologically relevant to the best of my knowledge.

        Parameters
        ----------
        other : VariantPosition
            The object to calculate adjacency to.

        Returns
        -------
        bool
            True if the positions describe adjacent bases in sequence space; else False.

        """
        if self.utr == other.utr:
            if self.intronic_position is None and other.intronic_position is None:
                return abs(self.position - other.position) == 1
            elif (
                self.position == other.position
            ):  # intronic positions can only be adjacent if they are relative to the same base
                if (
                    self.intronic_position is not None
                    and other.intronic_position is not None
                ):
                    return abs(self.intronic_position - other.intronic_position) == 1
                else:  # handle special case for first/last base of intron and corresponding first/last base of exon
                    return (
                        self.intronic_position == -1
                        or self.intronic_position == 1
                        or other.intronic_position == -1
                        or other.intronic_position == 1
                    )
            else:
                return False
        else:  # handle special case for last base of 5' utr and first base of non-UTR sequence
            return (self.position == -1 and other.position == 1) or (
                other.position == -1 and self.position == 1
            )



import unittest


class TestCombinePatterns(unittest.TestCase):
    @unittest.expectedFailure
    def test_something(self):
        self.assertEqual(True, False)


class TestRemoveNamedGroups(unittest.TestCase):
    @unittest.expectedFailure
    def test_something(self):
        self.assertEqual(True, False)


if __name__ == "__main__":
    unittest.main()



from fqfa.constants import DNA_BASES
from mavehgvs.patterns.util import combine_patterns, remove_named_groups
from mavehgvs.patterns.position import pos, pos_intron, pos_intron_utr

dna_nt: str = rf"[{''.join(DNA_BASES)}]"
"""str: Pattern matching any uppercase DNA base.

This does not include IUPAC ambiguity characters.
"""

dna_sub_c: str = rf"(?P<dna_sub_c>(?:(?P<position>{pos_intron_utr})(?P<ref>{dna_nt})>(?P<new>{dna_nt}))|(?P<equal>=))"
"""str: Pattern matching a DNA substitution with numeric, intronic, or UTR positions.
"""

dna_del_c: str = rf"(?P<dna_del_c>(?:(?:(?P<start>{pos_intron_utr})_(?P<end>{pos_intron_utr}))|(?P<pos>{pos_intron_utr}))del)"
"""str: Pattern matching a DNA deletion with numeric, intronic, or UTR positions.
"""

dna_dup_c: str = rf"(?P<dna_dup_c>(?:(?:(?P<start>{pos_intron_utr})_(?P<end>{pos_intron_utr}))|(?P<pos>{pos_intron_utr}))dup)"
"""str: Pattern matching a DNA duplication with numeric, intronic, or UTR positions.
"""

dna_ins_c: str = rf"(?P<dna_ins_c>(?P<start>{pos_intron_utr})_(?P<end>{pos_intron_utr})ins(?P<seq>{dna_nt}+))"
"""str: Pattern matching a DNA insertion with numeric, intronic, or UTR positions.
"""

dna_delins_c: str = rf"(?P<dna_delins_c>(?:(?:(?P<start>{pos_intron_utr})_(?P<end>{pos_intron_utr}))|(?P<pos>{pos_intron_utr}))delins(?P<seq>{dna_nt}+))"
"""str: Pattern matching a DNA deletion-insertion with numeric, intronic, or UTR positions.
"""

dna_sub_n: str = dna_sub_c.replace(pos_intron_utr, pos_intron).replace(
    "(?P<dna_sub_c>", "(?P<dna_sub_n>"
)
"""str: Pattern matching a DNA substitution with numeric or intron positions for non-coding variants.
"""

dna_del_n: str = dna_del_c.replace(pos_intron_utr, pos_intron).replace(
    "(?P<dna_del_c>", "(?P<dna_del_n>"
)
"""str: Pattern matching a DNA deletion with numeric or intron positions for non-coding variants.
"""

dna_dup_n: str = dna_dup_c.replace(pos_intron_utr, pos_intron).replace(
    "(?P<dna_dup_c>", "(?P<dna_dup_n>"
)
"""str: Pattern matching a DNA duplication with numeric or intron positions for non-coding variants.
"""

dna_ins_n: str = dna_ins_c.replace(pos_intron_utr, pos_intron).replace(
    "(?P<dna_ins_c>", "(?P<dna_ins_n>"
)
"""str: Pattern matching a DNA insertion with numeric or intron positions for non-coding variants.
"""

dna_delins_n: str = dna_delins_c.replace(pos_intron_utr, pos_intron).replace(
    "(?P<dna_delins_c>", "(?P<dna_delins_n>"
)
"""str: Pattern matching a DNA deletion-insertion with numeric or intron positions for non-coding variants.
"""

dna_sub_gmo: str = dna_sub_c.replace(pos_intron_utr, pos).replace(
    "(?P<dna_sub_c>", "(?P<dna_sub_gmo>"
)
"""str: Pattern matching a DNA substitution with only numeric positions for genomic-style variants.
"""

dna_del_gmo: str = dna_del_c.replace(pos_intron_utr, pos).replace(
    "(?P<dna_del_c>", "(?P<dna_del_gmo>"
)
"""str: Pattern matching a DNA deletion with only numeric positions for genomic-style variants.
"""

dna_dup_gmo: str = dna_dup_c.replace(pos_intron_utr, pos).replace(
    "(?P<dna_dup_c>", "(?P<dna_dup_gmo>"
)
"""str: Pattern matching a DNA duplication with only numeric positions for genomic-style variants.
"""

dna_ins_gmo: str = dna_ins_c.replace(pos_intron_utr, pos).replace(
    "(?P<dna_ins_c>", "(?P<dna_ins_gmo>"
)
"""str: Pattern matching a DNA insertion with only numeric positions for genomic-style variants.
"""

dna_delins_gmo: str = dna_delins_c.replace(pos_intron_utr, pos).replace(
    "(?P<dna_delins_c>", "(?P<dna_delins_gmo>"
)
"""str: Pattern matching a DNA deletion-insertion with only numeric positions for genomic-style variants.
"""

dna_variant_c: str = combine_patterns(
    [dna_sub_c, dna_del_c, dna_dup_c, dna_ins_c, dna_delins_c], None
)
"""str: Pattern matching any of the coding DNA variants.
"""

dna_variant_n: str = combine_patterns(
    [dna_sub_n, dna_del_n, dna_dup_n, dna_ins_n, dna_delins_n], None
)
"""str: Pattern matching any of the non-coding DNA variants.
"""

dna_variant_gmo: str = combine_patterns(
    [dna_sub_gmo, dna_del_gmo, dna_dup_gmo, dna_ins_gmo, dna_delins_gmo], None
)
"""str: Pattern matching any of the genomic-style DNA variants.
"""

dna_single_variant: str = rf"(?P<dna_c>c\.{dna_variant_c})|(?P<dna_n>n\.{dna_variant_n})|(?P<dna_gmo>[gmo]\.{dna_variant_gmo})"
"""str: Pattern matching any complete single DNA variant, including the prefix character.
"""

dna_multi_variant: str = rf"(?P<dna_c_multi>c\.\[{remove_named_groups(dna_variant_c)}(?:;{remove_named_groups(dna_variant_c)}){{1,}}\])|(?P<dna_n_multi>n\.\[{remove_named_groups(dna_variant_n)}(?:;{remove_named_groups(dna_variant_n)}){{1,}}\])|(?P<dna_gmo_multi>[gmo]\.\[{remove_named_groups(dna_variant_gmo)}(?:;{remove_named_groups(dna_variant_gmo)}){{1,}})\]"
"""str: Pattern matching any complete DNA multi-variant, including the prefix character.

Named capture groups have been removed from the variant patterns because of non-uniqueness.
Another applications of single-variant regular expressions is needed to recover the named groups from each individual
variant in the multi-variant.
"""



from mavehgvs.patterns.dna import dna_single_variant, dna_multi_variant
from mavehgvs.patterns.rna import rna_single_variant, rna_multi_variant
from mavehgvs.patterns.protein import pro_single_variant, pro_multi_variant

any_variant = rf"(?:(?P<target_id>[a-zA-Z0-9_.-]+):)?(?P<single_variant>{r'|'.join([dna_single_variant, rna_single_variant, pro_single_variant])})|(?P<multi_variant>{r'|'.join([dna_multi_variant, rna_multi_variant, pro_multi_variant])})"



import unittest

from mavehgvs.util import parse_variant_strings
from mavehgvs.variant import Variant


class TestParseVariantStrings(unittest.TestCase):
    def test_sets_error_strings_for_invalid_items(self) -> None:
        invalid_variant_strings = [
            "g.Glu27Trp",
            "p.27Glu>Trp",
            "p.122-6T>A",
            "G>A",
            "22G>A",
            "G.44del",
            "a.78+5_78+10del",
            "77dup",
            "n.Pro12_Gly18dup",
            "g.22_23insauc",
            "g.25_24del",
            "g.25_24ins",
            "r.43-6_595+12delinsctt",
            "x.=",
            "c.(=)",
        ]

        for s in invalid_variant_strings:
            with self.subTest(s=s):
                valid, invalid = parse_variant_strings([s])
                self.assertIsNone(valid[0])
                self.assertIsInstance(invalid[0], str)

    def test_sets_variant_for_valid_items(self) -> None:
        valid_variant_strings = [
            "p.Glu27Trp",
            "c.122-6T>A",
            "g.44del",
            "c.78+5_78+10del",
            "c.77dup",
            "p.Pro12_Gly18dup",
            "p.Ala12_Pro13insGlyProCys",
            "r.22_23insauc",
            "c.43-6_595+12delinsCTT",
            "p.Ile71_Cys80delinsSer",
            "p.=",
            "c.=",
            "p.(=)",
        ]

        for s in valid_variant_strings:
            with self.subTest(s=s):
                valid, invalid = parse_variant_strings([s])
                self.assertIsInstance(valid[0], Variant)
                self.assertIsNone(invalid[0])



import re
import itertools
from typing import Optional, Union, List, Tuple, Mapping, Any, Sequence, Dict, Generator

from fqfa.constants import AA_CODES

from mavehgvs.position import VariantPosition
from mavehgvs.patterns.combined import any_variant
from mavehgvs.exceptions import MaveHgvsParseError

__all__ = ["Variant"]

AA_3_TO_1 = {value: key for key, value in AA_CODES.items()}
"""Dict[str, str]: for converting three-letter amino acid codes to single-letter codes.
"""

class Variant:
    fullmatch = re.compile(any_variant, flags=re.ASCII).fullmatch
    """Callable[[str, int, int], Optional[Match[str]]]: fullmatch callable for parsing a single MAVE-HGVS variant
    
    Returns an :py:obj:`re.Match` object if the full string defines a valid MAVE-HGVS variant.
    Match groups in the result can be used to extract components of the variant.
    """

    VTYPES = (
        "sub",  # substitution
        "del",  # deletion
        "dup",  # duplication
        "ins",  # insertion
        "delins",  # deletion-insertion"
    )
    """Tuple[str]: variant type tags used in MAVE-HGVS patterns and variant type names.
    """

    def __init__(
        self,
        s: Union[str, Mapping[str, Any], Sequence[Mapping[str, Any]]],
        targetseq: Optional[str] = None,
        relaxed_ordering: bool = False,
    ):
        """Convert a MAVE-HGVS variant string into a corresponding object with named fields.

        Parameters
        ----------
        s : Union[str, Mapping[str, Any], Sequence[Mapping[str, Any]]]
            MAVE-HGVS variant string to convert into an object, dictionary type object containing key-value pairs
            corresponding to a MAVE-HGVS object, or list/tuple of dictionary type objects for a variant with
            multiple events.

        targetseq : Optional[str]
            If provided, the variant will be validated for agreement with this sequence.
            Target sequence validation is not supported for variants using the extended position syntax.

            This must be an amino acid sequence for protein variants or a nucleotide sequence for
            coding/noncoding/genomic variants.
            DNA and amino acid sequences should be in uppercase, RNA in lowercase.

        relaxed_ordering : bool
            If True, variants that do not observe the 3-prime rule for variant position ordering are allowed.
            The object representation will observe the 3-prime rule, so it may differ from the input string in this
            case.

        """
        if isinstance(s, str):  # variant string to parse
            variant_string = s
        elif isinstance(s, Mapping):  # dictionary-style single variant
            variant_string = self._variant_dictionary_to_string(s, include_prefix=True)
        elif isinstance(s, Sequence):  # dictionary-style multi-variant
            try:
                all_prefixes = [v["prefix"] for v in s]
            except KeyError:
                raise MaveHgvsParseError("variant dictionary missing required keys")
            if len(set(all_prefixes)) != 1:
                raise MaveHgvsParseError(
                    "cannot combine variants with different prefixes"
                )
            variant_string = f"{s[0]['prefix']}.[{';'.join(self._variant_dictionary_to_string(v, include_prefix=False) for v in s)}]"
        else:
            raise ValueError("can only create Variants from string or Mapping objects")

        variant_match = self.fullmatch(variant_string)
        if variant_match is None:
            raise MaveHgvsParseError("failed regular expression validation")
        else:
            match_dict = variant_match.groupdict()

            # set target id if present
            if match_dict["target_id"] is not None:
                self._target_id = match_dict["target_id"]
            else:
                self._target_id = None

            # set prefix and determine if this is a multi-variant
            if match_dict["single_variant"] is not None:
                self.variant_count = 1
                self._prefix = match_dict["single_variant"][0]
            elif match_dict["multi_variant"] is not None:
                self.variant_count = len(variant_string.split(";"))
                self._prefix = match_dict["multi_variant"][0]
            else:  # pragma: no cover
                raise ValueError("invalid match type")

            if self.variant_count == 1:
                self._variant_types, self._positions, self._sequences = self._process_string_variant(
                    match_dict, relaxed_ordering=relaxed_ordering
                )
            elif self.variant_count > 1:
                self._variant_types = list()
                self._positions = list()
                self._sequences = list()

                # format each individual variant event as a single variant and parse it
                for variant_substring in match_dict["multi_variant"][3:-1].split(";"):
                    groupdict = self.fullmatch(
                        f"{self._prefix}.{variant_substring}"
                    ).groupdict()
                    vt, p, s = self._process_string_variant(
                        groupdict, relaxed_ordering=relaxed_ordering
                    )
                    if p is None:  # only the case for target-identical variants
                        raise MaveHgvsParseError(
                            "multi-variants cannot contain target-identical variants"
                        )

                    self._variant_types.append(vt)
                    self._positions.append(p)
                    self._sequences.append(s)

                # ensure that multiple variants aren't defined for the same positions
                for vp1, vp2 in itertools.combinations(self._positions, 2):
                    if isinstance(vp1, VariantPosition) and isinstance(
                        vp2, VariantPosition
                    ):  # both single position
                        if vp1 == vp2:
                            raise MaveHgvsParseError(
                                "multi-variant has multiple changes for the same position"
                            )
                    elif isinstance(vp1, VariantPosition) and isinstance(vp2, Tuple):
                        if vp2[0] <= vp1 <= vp2[1]:
                            raise MaveHgvsParseError(
                                "multi-variant has overlapping changes"
                            )
                    elif isinstance(vp1, Tuple) and isinstance(vp2, VariantPosition):
                        if vp1[0] <= vp2 <= vp1[1]:
                            raise MaveHgvsParseError(
                                "multi-variant has overlapping changes"
                            )
                    elif isinstance(vp1, Tuple) and isinstance(vp2, Tuple):
                        if (
                            vp1[0] <= vp2[0] <= vp1[1]
                            or vp1[0] <= vp2[1] <= vp1[1]
                            or vp2[0] <= vp1[0] <= vp2[1]
                            or vp2[0] <= vp1[1] <= vp2[1]
                        ):
                            raise MaveHgvsParseError(
                                "multi-variant has overlapping changes"
                            )
                    else:  # pragma: no cover
                        raise ValueError("invalid position type")

                # re-order variants and validate
                def sort_key(x):
                    if isinstance(x[1], VariantPosition):
                        return x[1]
                    elif isinstance(x[1], Tuple):
                        return x[1][0]
                    else:
                        raise ValueError("invalid position type")

                variant_list = list(self.variant_tuples())
                ordered_list = sorted(variant_list, key=sort_key)
                if variant_list != ordered_list:
                    if relaxed_ordering:
                        self._variant_types = [x[0] for x in ordered_list]
                        self._positions = [x[1] for x in ordered_list]
                        self._sequences = [x[2] for x in ordered_list]
                    else:
                        raise MaveHgvsParseError("multi-variants not in sorted order")

            else:  # pragma: no cover
                raise ValueError("invalid variant count")

        if targetseq is not None:
            for vtype, pos, seq in self.variant_tuples():
                if vtype == "sub":
                    if self._prefix == "p":
                        ref = AA_3_TO_1[seq[0]]
                    else:
                        ref = seq[0]
                    self._target_validate_substitution(pos, ref, targetseq)
                elif vtype in ("ins", "del", "dup", "delins"):
                    self._target_validate_indel(pos, targetseq)

    def variant_tuples(
        self
    ) -> Generator[
        Tuple[
            str,
            Optional[Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]],
            Optional[Union[str, Tuple[str, str]]],
        ],
        None,
        None,
    ]:
        """Generator that yields tuples containing the variant components.

        Yields
        ------
        Tuple
            Tuple of the variant type, position(s), and sequence(s) for each element in the variant.

        """
        if self.is_multi_variant():
            for vtype, pos, seq in zip(
                self._variant_types, self._positions, self._sequences
            ):
                yield vtype, pos, seq
        else:
            yield self._variant_types, self._positions, self._sequences

    def _process_string_variant(
        self, match_dict: Dict[str, str], relaxed_ordering: bool
    ) -> Tuple[
        str,
        Optional[Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]],
        Optional[Union[str, Tuple[str, str]]],
    ]:
        """Process the match dictionary from a single variant into its components.

        Parameters
        ----------
        match_dict : Dict[str, str]
            Match dictionary from the MAVE-HGVS regular expression.
        relaxed_ordering : bool
            If True, variants that do not observe the 3-prime rule for variant position ordering are allowed.

        Returns
        -------
        Tuple[str, Optional[Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]], Optional[Union[str, Tuple[str, str]]]]
            Returns a 3-tuple containing the variant type, optional position (or start/end positions),
            and optional before/after substitution sequences or inserted sequence.

        """
        variant_type = None
        positions = None
        sequences = None

        # determine which named groups to check
        if self._prefix == "p":
            pattern_group_tuples = [(f"pro_{t}", t) for t in self.VTYPES]
        elif self._prefix == "r":
            pattern_group_tuples = [(f"rna_{t}", t) for t in self.VTYPES]
        elif self._prefix in "cn":
            pattern_group_tuples = [(f"dna_{t}_{self._prefix}", t) for t in self.VTYPES]
        elif self._prefix in "gmo":
            pattern_group_tuples = [(f"dna_{t}_gmo", t) for t in self.VTYPES]
        else:  # pragma: no cover
            raise ValueError("unexpected prefix")

        # set the variant type
        vtype_set = False
        pattern_group = None
        for pg, vtype in pattern_group_tuples:
            if match_dict[pg] is not None:
                if vtype_set:  # pragma: no cover
                    raise ValueError(f"ambiguous match: '{pg}' and '{pattern_group}'")
                variant_type = vtype
                pattern_group = pg
                vtype_set = True

        # set the position and sequence
        if variant_type == "sub":
            if (
                match_dict[f"{pattern_group}_equal"] is not None
            ):  # special case for target identity
                sequences = match_dict[f"{pattern_group}_equal"]
            elif match_dict[f"pro_sub_equal_sy"] is not None:
                sequences = match_dict[f"pro_sub_equal_sy"]
            else:
                positions = VariantPosition(match_dict[f"{pattern_group}_position"])
                if self._prefix == "p":
                    sequences = (
                        positions.amino_acid,
                        match_dict[f"{pattern_group}_new"],
                    )
                elif self._prefix in "gmo" "cn" "r":
                    sequences = (
                        match_dict[f"{pattern_group}_ref"],
                        match_dict[f"{pattern_group}_new"],
                    )
                else:  # pragma: no cover
                    raise ValueError("unexpected prefix")
        elif variant_type in ("del", "dup", "ins", "delins"):
            # set position
            if (
                match_dict.get(f"{pattern_group}_pos") is not None
            ):  # use get() since ins pattern doesn't have pos
                positions = VariantPosition(match_dict[f"{pattern_group}_pos"])
            else:
                positions = (
                    VariantPosition(match_dict[f"{pattern_group}_start"]),
                    VariantPosition(match_dict[f"{pattern_group}_end"]),
                )
                # extra validation on positions
                if positions[0] >= positions[1]:
                    if relaxed_ordering:
                        positions = (positions[1], positions[0])
                    else:
                        raise MaveHgvsParseError(
                            "start position must be before end position"
                        )
                if variant_type == "ins":
                    if not positions[0].is_adjacent(positions[1]):
                        raise MaveHgvsParseError("insertion positions must be adjacent")

            # set sequence if needed
            if variant_type in ("ins", "delins"):
                sequences = match_dict[f"{pattern_group}_seq"]

        return variant_type, positions, sequences

    # TODO: API documentation for the dictionary objects
    @staticmethod
    def _variant_dictionary_to_string(
        vdict: Mapping[str, Any], include_prefix: bool
    ) -> str:
        """Convert a match dictionary from a single variant into a string for further validation.

        This method performs minimal validation of the values provided in the input, and instead converts it into a
        variant string that is validated using the regular expression based validators.

        Parameters
        ----------
        vdict : Mapping[str, Any]
            Key-value pairs describing a single variant.
        include_prefix: bool
            If True, the variant prefix and '.' will be included in the string; else it is omitted (for use with
            multi-variants).

        Returns
        -------
        str
            A string representing this variant.

        Raises
        ------
        MaveHgvsParseError
            If the dictionary does not have a valid set of keys.

        """
        try:
            variant_type = vdict["variant_type"]
        except KeyError:
            raise MaveHgvsParseError("variant dictionary missing required keys")

        if variant_type == "sub":
            if sorted(vdict.keys()) != sorted(
                ["variant_type", "prefix", "position", "target", "variant"]
            ):
                raise MaveHgvsParseError("variant dictionary contains invalid keys")
            if vdict["prefix"] == "p":
                variant_string = (
                    f"{vdict['target']}{vdict['position']}{vdict['variant']}"
                )
            else:
                variant_string = (
                    f"{vdict['position']}{vdict['target']}>{vdict['variant']}"
                )
        elif variant_type in ("del", "dup"):
            if sorted(vdict.keys()) != sorted(
                ["variant_type", "prefix", "start_position", "end_position"]
            ):
                raise MaveHgvsParseError("variant dictionary contains invalid keys")
            if vdict["start_position"] == vdict["end_position"]:
                variant_string = f"{vdict['start_position']}{variant_type}"
            else:
                variant_string = (
                    f"{vdict['start_position']}_{vdict['end_position']}{variant_type}"
                )
        elif variant_type in ("ins", "delins"):
            if sorted(vdict.keys()) != sorted(
                ["variant_type", "prefix", "start_position", "end_position", "sequence"]
            ):
                raise MaveHgvsParseError("variant dictionary contains invalid keys")
            variant_string = f"{vdict['start_position']}_{vdict['end_position']}{variant_type}{vdict['sequence']}"
        else:
            raise MaveHgvsParseError("invalid variant type")

        if include_prefix:
            return f"{vdict['prefix']}.{variant_string}"
        else:
            return variant_string

    def __repr__(self) -> str:
        """The object representation is equivalent to the input string.

        Returns
        -------
        str
            The object representation.

        """

        def format_variant(
            vtype: str,
            pos: Union[VariantPosition, Tuple[VariantPosition, VariantPosition]],
            seq: Optional[Union[str, Tuple[str, str]]],
        ) -> str:
            """Helper function for building variant strings.

            Parameters
            ----------
            vtype : str
                The variant type, as described by :py:obj:`Variant.__vtypes`
            pos : Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]
                The position or pair of positions describing the variant.
            seq : Optional[Union[str, Tuple[str, str]]]
                The sequence or pair of sequences describing the variant.
                Only used for substitions, insertions, and deletion-insertions.

            Returns
            -------
            str
                A string representing this variant element.

            """
            if vtype == "sub":
                if self._prefix == "p":  # protein variant
                    return f"{pos}{seq[1]}"
                else:  # nucleotide variant
                    return f"{pos}{seq[0]}>{seq[1]}"
            elif vtype in ("del", "dup"):
                if isinstance(pos, tuple):
                    return f"{pos[0]}_{pos[1]}{vtype}"
                else:
                    return f"{pos}{vtype}"
            elif vtype in ("ins", "delins"):
                if isinstance(pos, tuple):
                    return f"{pos[0]}_{pos[1]}{vtype}{seq}"
                else:
                    return f"{pos}{vtype}{seq}"
            else:  # pragma: no cover
                raise ValueError("invalid variant type")

        if self._target_id is not None:
            prefix = f"{self._target_id}:{self._prefix}"
        else:
            prefix = f"{self._prefix}"

        if self.is_target_identical():
            return f"{prefix}.{self._sequences}"
        else:
            elements = [format_variant(*t) for t in self.variant_tuples()]
            if self.is_multi_variant():
                return f"{prefix}.[{';'.join(elements)}]"
            else:
                return f"{prefix}.{elements[0]}"

    @staticmethod
    def _target_validate_substitution(
        pos: VariantPosition, ref: str, target: str
    ) -> None:
        """Determine whether the target portion of a substitution matches the target sequence.

        Note that variants using extended syntax cannot be validated with this method.
        If an extended syntax variant is encountered, it will be interpreted as valid/matching.

        Parameters
        ----------
        pos : VariantPosition
            Position of the substitution.
        ref : str
            Reference base or amino acid from the variant.
        target : str
            Target sequence. This must be an amino acid sequence for protein variants or a nucleotide sequence
            for coding/noncoding/genomic variants.
            RNA sequences should be in lowercase, DNA sequences should be in uppercase.

        Returns
        -------
        None

        Raises
        ------
        MaveHgvsParseError
            If the reference base or amino acid does not match the target at the given position
        MaveHgvsParseError
            If the position is outside the bounds of the target.

        """
        if pos.is_extended():
            return
        elif pos.position > len(target):
            raise MaveHgvsParseError("variant coordinate out of bounds")
        elif target[pos.position - 1] != ref:
            raise MaveHgvsParseError("substitution reference does not match target")
        else:
            return

    @staticmethod
    def _target_validate_indel(
        pos: Union[VariantPosition, Tuple[VariantPosition, VariantPosition]],
        target: str,
    ) -> None:
        """Determine whether indel coordinates are valid for the target sequence.

        Note that variants using extended syntax cannot be validated with this method.
        If an extended syntax variant is encountered, it will be interpreted as valid/matching.

        Parameters
        ----------
        pos : Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]
            Single variant position or start/end tuple for the indel.
        target : str
            Target sequence. This must be an amino acid sequence for protein variants or a nucleotide sequence
            for coding/noncoding/genomic variants.

        Returns
        -------
        None

        Raises
        ------
        MaveHgvsParseError
            If the indel coordinates are outside the target bounds.

        """
        if isinstance(pos, tuple):
            start, end = pos
            if start.is_extended() or end.is_extended():
                return
            elif start.position > len(target) or end.position > len(target):
                raise MaveHgvsParseError("variant coordinate out of bounds")
        elif pos.position > len(target):
            raise MaveHgvsParseError("substitution coordinate out of bounds")
        else:
            return

    def is_target_identical(self) -> bool:
        """Return whether the variant describes the "wild-type" sequence or is the special synonymous variant.

        This is the variant described with only the equals sign (e.g. ``c.=``)
        or the uncertain equals protein variant (e.g. ``p.(=)``).

        Returns
        -------
        bool
            True if this variant describes the wild-type or target sequence; else False.

        """
        return self._positions is None

    def is_multi_variant(self) -> bool:
        """Return whether the variant is a multi-variant.

        A multi-variant is a single variant describing multiple events enclosed in '[]'.
        Multi-variants are referred to as alleles in the HGVS standard.

        Returns
        -------
        bool
            True if the variant is a multi-variant; else False.

        """
        return self.variant_count > 1

    @property
    def prefix(self) -> str:
        """The single-letter prefix for this variant.

        Returns
        -------
        str
            Single-letter prefix corresponding to the sequence type.

            See the following table for sequence type prefixes and their meanings:

            .. csv-table::
               :file: ../docs/prefix.csv
               :header: "Prefix", "Description"
               :widths: 5, 20

        """
        return self._prefix

    @property
    def variant_type(self) -> Union[str, List[str]]:
        """The type for this variant.

        Valid variant types are:

        * ``'sub'`` for substitutions
        * ``'del'`` for deletions
        * ``'dup'`` for duplications
        * ``'ins'`` for insertions
        * ``'delins'`` for deletion-insertions

        Returns
        -------
        Union[str, List[str]]
            String containing the variant type. Returns a list of strings for a multi-variant.

        """
        return self._variant_types

    def uses_extended_positions(self) -> bool:
        """Return whether the variant uses the extended position notation to describe intronic or UTR positions.

        Examples of variants using the extended position notation include:

        * c.122-6T>A
        * r.*33a>c
        * c.43-6_595+12delinsCTT

        This should always be false for variants with a genomic or protein prefix, as variants with these prefixes
        cannot use positions relative to a transcript under the MAVE-HGVS specification.

        Returns
        -------
        bool
            True if the variant (or any of the individual variants for a multi-variant) uses the extended position
            notation.

        """
        if self.is_multi_variant():
            all_positions = list()
            for p in self.positions:
                if isinstance(p, tuple):
                    all_positions.extend(p)
                else:
                    all_positions.append(p)
            return any(p.is_extended() for p in all_positions)
        else:
            if self._positions is None:  # special case for target identity
                return False
            elif isinstance(self.positions, tuple):
                return any(p.is_extended() for p in self.positions)
            else:
                return self.positions.is_extended()

    @property
    def positions(
        self
    ) -> Optional[
        Union[
            VariantPosition,
            Tuple[VariantPosition, VariantPosition],
            List[Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]],
        ]
    ]:
        """The variant position as a single position or tuple containing start and end positions.

        Each position is an instance of :py:class:`mavehgvs.position.VariantPosition`.

        Returns
        -------
        Union[VariantPosition, Tuple[VariantPosition, VariantPosition], List[Union[VariantPosition, Tuple[VariantPosition, VariantPosition]]]]
            Variant position or tuple of start/end positions.
            Returns a list of positions or start/end tuples for a multi-variant.

        """
        return self._positions

    @property
    def sequence(
        self
    ) -> Optional[
        Union[str, Tuple[str, str], List[Optional[Union[str, Tuple[str, str]]]]]
    ]:
        """The sequence portion of the variant.

        This can be a tuple of target and new bases for a substitution, a single sequence for insertions or
        deletion-insertions, or the "=" character for variants that are identical to the target sequence.

        Returns
        -------
        Union[str, Tuple[str, str], List[Optional[Union[str, Tuple[str, str]]]]]]
            Tuple of ref/new bases for substitutions, string containing inserted sequence, or the "=" character.
            Returns None if the variant does not have a sequence component (deletion or duplication).
            Returns a list for a multi-variant, which may contain None values for deletions or duplications.

        """
        return self._sequences

    @property
    def target_id(self) -> Optional[str]:
        """The target identifier for the variant (if applicable).

        The target identifier precedes the prefix and is followed by a ``:``.
        For example in ``NM_001130145.3:c.832C>T`` the target identifier is "NM_001130145.3".

        Returns
        -------
        Optional[str]
            The target identifier, or None if it is not set.

        """
        return self._target_id



"""Utility functions for working with mavehgvs regex pattern strings.
"""

import re
from typing import Sequence, Optional


def combine_patterns(patterns: Sequence[str], groupname: Optional[str] = None) -> str:
    """Combine multiple pattern strings into a single pattern string.

    Because multiple identical group names are not allowed in a pattern, the resulting object renames all named match
    groups such they are prefixed with the first match group name in the pattern. For example,
    ``(?P<substitution>(?P<position>[1-9][0-9]*)...`` becomes
    ``(?P<substitution>(?P<substitution_position>[1-9][0-9]*)...``.

    The function assumes that all input patterns are enclosed in parentheses.

    Parameters
    ----------
    patterns : Sequence[str]
        Sequence of pattern strings to combine.

    groupname : Optional[str]
        Name for the capture group surrounding the resulting pattern. If this is None, a non-capturing group will be
        used instead.

    Returns
    -------
    str
        Pattern string that matches any of the input patterns. Match groups are renamed as described above to attempt
        to ensure uniqueness across the combined pattern.

    """
    tag_re = re.compile(r"\(\?P<(\w+)>")
    stripped_patterns = list()
    for p in patterns:
        tags = list(tag_re.finditer(p))
        prefix = f"{tags[0].group(1)}_"
        new_p = p
        for t in tags[:0:-1]:
            start, end = t.span(1)
            new_p = "".join((new_p[:start], prefix, new_p[start:]))
        stripped_patterns.append(new_p)
    if groupname is None:
        combined = rf"(?:{r'|'.join(stripped_patterns)})"
    else:
        combined = rf"(?P<{groupname}>{r'|'.join(stripped_patterns)})"

    return combined


def remove_named_groups(pattern: str, noncapturing: bool = True) -> str:
    """Function that replaces named match groups in a regular expression pattern.

    Named groups are replaced with either regular parentheses or non-capturing parentheses.

    Parameters
    ----------
    pattern : str
        The pattern string to strip match groups from.

    noncapturing : bool
        If True, the named grouping parentheses are replaced by non-capturing parentheses.
        If False, regular parentheses are used.

    Returns
    -------
    str
        The pattern string without named match groups.

    """
    if noncapturing:
        new_parens = "(?:"
    else:
        new_parens = "("

    return re.sub(r"\(\?P<\w+>", new_parens, pattern)



import unittest
import itertools
import random
from mavehgvs.position import VariantPosition
from mavehgvs.exceptions import MaveHgvsParseError


class TestObjectCreation(unittest.TestCase):
    def test_position_only(self) -> None:
        v = VariantPosition("8")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (8, None, None, None),
        )

        v = VariantPosition("92380")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (92380, None, None, None),
        )

    def test_amino_acid(self) -> None:
        v = VariantPosition("Gly8")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (8, "Gly", None, None),
        )

        v = VariantPosition("Cys92380")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (92380, "Cys", None, None),
        )

    def test_invalid_strings(self) -> None:
        position_strings = (
            "08",
            "+12",
            "*-99",
            "A",
            "TCGA",
            "g",
            "*",
            "-",
            "+",
            "**6",
            "800 + 12",
            "-12*5",
            "Glu-12",
            "*5Trp",
            "Xyz12",
            "ALA12",
        )
        for s in position_strings:
            with self.subTest(s=s):
                with self.assertRaises(MaveHgvsParseError):
                    VariantPosition(s)

    def test_utr(self) -> None:
        v = VariantPosition("*8")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (8, None, None, True),
        )

        v = VariantPosition("-80")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (-80, None, None, True),
        )

    def test_intron(self) -> None:
        v = VariantPosition("122-6")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (122, None, -6, None),
        )

        v = VariantPosition("78+10")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr), (78, None, 10, None)
        )

    def test_utr_intron(self) -> None:
        v = VariantPosition("*89+67")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr), (89, None, 67, True)
        )

        v = VariantPosition("-127+6")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (-127, None, 6, True),
        )

        v = VariantPosition("*73-105")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (73, None, -105, True),
        )

        v = VariantPosition("-45-1")
        self.assertTupleEqual(
            (v.position, v.amino_acid, v.intronic_position, v.utr),
            (-45, None, -1, True),
        )


class TestObjectRepresentation(unittest.TestCase):
    def test_repr(self) -> None:
        position_strings = (
            "8",
            "92380",
            "*8",
            "-80",
            "122-6",
            "78+10",
            "*89+67",
            "-127+6",
            "*73-105",
            "-45-1",
            "Cys234",
            "Ala9",
        )
        for s in position_strings:
            with self.subTest(s=s):
                v = VariantPosition(s)
                self.assertEqual(s, repr(v))


# TODO: add amino acid variants
class TestComparisons(unittest.TestCase):
    def setUp(self) -> None:
        sorted_position_strings = (
            "-45-1",
            "-12",
            "8",
            "99",
            "99+88",
            "99+122",
            "100-12",
            "100",
            "101",
            "202-12",
            "202-1",
            "202",
            "*1",
            "*73-105",
        )

        self.sorted_variants = [VariantPosition(p) for p in sorted_position_strings]

        # pairwise itertools recipe
        a, b = itertools.tee(self.sorted_variants)
        next(b, None)
        self.sorted_variant_pairs = zip(a, b)

    def test_eq(self) -> None:
        for v in self.sorted_variants:
            with self.subTest(v=v):
                self.assertEqual(v, v)

    def test_ne(self) -> None:
        for v1, v2 in self.sorted_variant_pairs:
            with self.subTest(v1=v1, v2=v2):
                self.assertNotEqual(v1, v2)

    def test_lt(self) -> None:
        for v1, v2 in self.sorted_variant_pairs:
            with self.subTest(v1=v1, v2=v2):
                self.assertLess(v1, v2)

    def test_sorting(self) -> None:
        for _ in range(10):
            with self.subTest():
                shuffled_variants = self.sorted_variants.copy()
                while shuffled_variants == self.sorted_variants:
                    random.shuffle(shuffled_variants)
                self.assertListEqual(self.sorted_variants, sorted(shuffled_variants))


# TODO: add amino acid variants
class TestAdjacency(unittest.TestCase):
    def test_adjacent_pairs(self) -> None:
        adjacent_pairs = (
            ("-45-2", "-45-1"),
            ("-45-1", "-45"),
            ("-12", "-13"),
            ("-1", "1"),
            ("8", "9"),
            ("202-1", "202"),
            ("99", "99+1"),
            ("99+88", "99+89"),
            ("100-12", "100-11"),
            ("100", "101"),
            ("*1", "*2"),
            ("*73-1", "*73"),
        )
        for s1, s2 in adjacent_pairs:
            v1 = VariantPosition(s1)
            v2 = VariantPosition(s2)
            with self.subTest(v1=v1, v2=v2):
                self.assertTrue(v1.is_adjacent(v2))
            with self.subTest(v1=v1, v2=v2):
                self.assertTrue(v2.is_adjacent(v1))

    def test_not_adjacent_to_self(self) -> None:
        position_strings = (
            "-45-1",
            "-12",
            "8",
            "99",
            "99+88",
            "99+122",
            "100-12",
            "100",
            "103",
            "202-12",
            "202-1",
            "205",
            "*1",
            "*12",
            "*73-105",
        )
        variants = [VariantPosition(s) for s in position_strings]
        for v in variants:
            with self.subTest(v=v):
                self.assertFalse(v.is_adjacent(v))

    def test_non_adjacent_pairs(self) -> None:
        position_strings = (
            "-45-1",
            "-12",
            "8",
            "99",
            "99+88",
            "99+122",
            "100-12",
            "103",
            "202-12",
            "202-1",
            "205",
            "*1",
            "*12",
            "*73-105",
        )
        variants = [VariantPosition(s) for s in position_strings]

        for v1, v2 in itertools.permutations(variants, 2):
            with self.subTest(v1=v1, v2=v2):
                self.assertFalse(v1.is_adjacent(v2))


if __name__ == "__main__":
    unittest.main()
